import { render } from "ink";
import Table from "ink-table";
import React from "react";
import { fetchResult } from "../cfetch";
import { formatBytes, formatTimeAgo } from "../formatTimeAgo";
import { requireAuth } from "../user";
import { getDatabaseByName } from "./list";
import type { Backup, Database } from "./types";
import type { ArgumentsCamelCase, Argv } from "yargs";

type BackupListArgs = { name: string };

export function ListOptions(yargs: Argv): Argv<BackupListArgs> {
	return yargs.positional("name", {
		describe: "The name of the DB",
		type: "string",
		demandOption: true,
	});
}

export async function ListHandler({
	name,
}: ArgumentsCamelCase<BackupListArgs>): Promise<void> {
	const accountId = await requireAuth({});
	const db: Database = await getDatabaseByName(accountId, name);

	const backups: Backup[] = await listBackups(accountId, db.uuid);
	render(
		<Table
			data={backups}
			columns={["created_at", "id", "num_tables", "size"]}
		></Table>
	);
}

export const listBackups = async (
	accountId: string,
	uuid: string
): Promise<Array<Backup>> => {
	const json: Backup[] = await fetchResult(
		`/accounts/${accountId}/d1/database/${uuid}/backup`,
		{}
	);
	const results: Record<string, Backup> = {};

	json
		// First, convert created_at to a Date
		.map((backup) => ({
			...backup,
			created_at: new Date(backup.created_at),
		}))
		// Then, sort descending based on created_at
		.sort((a, b) => +b.created_at - +a.created_at)
		// then group_by their human-readable timestamp i.e. "2 days ago"
		// (storing only the first of each group)
		// and replace the Date version with this new human-readable one
		.forEach((backup) => {
			const timeAgo = formatTimeAgo(backup.created_at);
			if (!results[timeAgo]) {
				results[timeAgo] = {
					...backup,
					created_at: timeAgo,
					size: formatBytes(backup.file_size),
				};
			}
		});

	// Take advantage of JS objects' sorting to return the newest backup of a certain age
	return Object.values(results);
};

// export function CreateOptions(yargs: Argv): Argv<BackupListArgs> {
//   return ListOptions(yargs)
// }
//
// export async function CreateHandler({
//   name,
// }: ArgumentsCamelCase<BackupListArgs>): Promise<void> {
//   const accountId = await requireAuth({});
//   const db: Database = await getDatabaseByName(accountId, name);
//
//   const backups: Backup[] = await createBackup(accountId, db.uuid);
//   render(
//     <Table
//       data={backups}
//       columns={["created_at", "id", "num_tables", "size"]}
//     ></Table>
//   );
// }
//
// export const createBackup = async (
//   accountId: string,
//   uuid: string
// ): Promise<Array<Backup>> => {
//
//   const db: Database = await fetchResult(`/accounts/${accountId}/d1/database`, {
//     method: "POST",
//     headers: {
//       "Content-Type": "application/json",
//     },
//     body: JSON.stringify({
//       name,
//     }),
//   });
// };
