import fs from "node:fs/promises";
import { render } from "ink";
import Table from "ink-table";
import React from "react";
import { fetchResult } from "../cfetch";
import { requireAuth } from "../user";
import { getDatabaseByName } from "./list";
import type { Database } from "./types";
import type { ArgumentsCamelCase, Argv } from "yargs";

type ExecuteArgs = {
	name: string;
	file?: string;
	command?: string;
};

type QueryResult = {
	results: Record<string, string | number | boolean>[];
	success: boolean;
	duration: number;
};

export function Options(yargs: Argv): Argv<ExecuteArgs> {
	return yargs
		.positional("name", {
			describe: "The name of the DB",
			type: "string",
			demandOption: true,
		})
		.option("file", {
			describe: "A .sql file to injest",
			type: "string",
		})
		.option("command", {
			describe: "A single SQL statement to execute",
			type: "string",
		});
}

export async function Handler({
	name,
	file,
	command,
}: ArgumentsCamelCase<ExecuteArgs>): Promise<void> {
	if (!file && !command)
		return console.error(`Error: must provide --command or --file.`);
	if (file && command)
		return console.error(`Error: can't provide both --command and --file.`);

	// Only multi-queries are working atm, so throw down a little extra one.
	const sql = file ? await fs.readFile(file, "utf8") : command;

	const accountId = await requireAuth({});
	const db: Database = await getDatabaseByName(accountId, name);
	console.log(`Executing on ${name} (${db.uuid}):`);

	const response: QueryResult[] = await fetchResult(
		`/accounts/${accountId}/d1/database/${db.uuid}/query`,
		{
			method: "POST",
			headers: {
				"Content-Type": "application/json",
			},
			body: JSON.stringify({ sql }),
		}
	);

	for (let i = 0; i < response.length; i++) {
		const { results, duration } = response[i];

		if (Array.isArray(results) && results.length > 0) {
			render(<Table data={results}></Table>);
		} else {
			console.log(`Executed command in ${duration}ms.`);
		}
	}
}
